import asyncio
import json
import logging
import logging.config
import os
import re
import socket
import ssl
import sys
import time
from functools import partial
from pathlib import Path
from pprint import pprint
from typing import Any, AsyncIterator, MutableMapping, Tuple

import aiohttp_cors
import aiotools
import click
import jinja2
import tomli
import uvloop
import yarl
from aiohttp import web
from redis.asyncio import Redis
from setproctitle import setproctitle

from ai.backend.client.config import APIConfig
from ai.backend.client.exceptions import BackendAPIError, BackendClientError
from ai.backend.client.session import AsyncSession as APISession
from ai.backend.common.web.session import extra_config_headers, get_session
from ai.backend.common.web.session import setup as setup_session
from ai.backend.common.web.session.redis_storage import RedisStorage

from . import __version__, user_agent
from .config import config_iv
from .logging import BraceStyleAdapter
from .proxy import decrypt_payload, web_handler, web_plugin_handler, websocket_handler
from .template import toml_scalar

log = BraceStyleAdapter(logging.getLogger("ai.backend.web.server"))

cache_patterns = {
    r"\.(?:manifest|appcache|html?|xml|json|ini|toml)$": {
        "Cache-Control": "no-store",
    },
    r"(?:backend.ai-webui.js)$": {
        "Cache-Control": "no-store",
    },
    r"\.(?:jpg|jpeg|gif|png|ico|cur|gz|svg|svgz|mp4|ogg|ogv|webm|htc|woff|woff2)$": {
        "Cache-Control": "max-age=259200, public",
    },
    r"\.(?:css|js)$": {
        "Cache-Control": "max-age=86400, public, must-revalidate, proxy-revalidate",
    },
    r"\.(?:py|log?|txt)$": {
        "Cache-Control": "no-store",
    },
}
_cache_patterns = {re.compile(k): v for k, v in cache_patterns.items()}


def apply_cache_headers(response: web.StreamResponse, path: str) -> web.StreamResponse:
    for regex, headers in _cache_patterns.items():
        mo = regex.search(path)
        if mo is not None:
            response.headers.update(headers)
            break
    return response


async def static_handler(request: web.Request) -> web.StreamResponse:
    request_path = request.match_info["path"]
    static_path = request.app["config"]["service"]["static_path"]
    file_path = (static_path / request_path).resolve()
    try:
        file_path.relative_to(static_path)
    except (ValueError, FileNotFoundError):
        return web.HTTPNotFound(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/generic-not-found",
                    "title": "Not Found",
                }
            ),
            content_type="application/problem+json",
        )
    if file_path.is_file():
        return apply_cache_headers(web.FileResponse(file_path), request_path)
    return web.HTTPNotFound(
        text=json.dumps(
            {
                "type": "https://api.backend.ai/probs/generic-not-found",
                "title": "Not Found",
            }
        ),
        content_type="application/problem+json",
    )


async def config_ini_handler(request: web.Request) -> web.Response:
    config = request.app["config"]
    scheme = config["service"]["force_endpoint_protocol"]
    if scheme is None:
        scheme = request.scheme
    j2env: jinja2.Environment = request.app["j2env"]
    tpl = j2env.get_template("config_ini.toml.j2")
    config_content = tpl.render(
        {
            "endpoint_url": f"{scheme}://{request.host}",  # must be absolute
            "config": config,
        }
    )
    return web.Response(text=config_content, content_type="text/plain")


async def config_toml_handler(request: web.Request) -> web.Response:
    config = request.app["config"]
    scheme = config["service"]["force_endpoint_protocol"]
    if scheme is None:
        scheme = request.scheme
    j2env: jinja2.Environment = request.app["j2env"]
    tpl = j2env.get_template("config.toml.j2")
    config_content = tpl.render(
        {
            "endpoint_url": f"{scheme}://{request.host}",  # must be absolute
            "config": config,
        }
    )
    return web.Response(text=config_content, content_type="text/plain")


async def console_handler(request: web.Request) -> web.StreamResponse:
    request_path = request.match_info["path"]
    config = request.app["config"]
    static_path = config["service"]["static_path"]
    file_path = (static_path / request_path).resolve()
    # SECURITY: only allow reading files under static_path
    try:
        file_path.relative_to(static_path)
    except (ValueError, FileNotFoundError):
        return web.HTTPNotFound(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/generic-not-found",
                    "title": "Not Found",
                }
            ),
            content_type="application/problem+json",
        )
    if file_path.is_file():
        return apply_cache_headers(web.FileResponse(file_path), request_path)
    # Fallback to index.html to support the URL routing for single-page application.
    return apply_cache_headers(web.FileResponse(static_path / "index.html"), "index.html")


async def login_check_handler(request: web.Request) -> web.Response:
    session = await get_session(request)
    authenticated = bool(session.get("authenticated", False))
    public_data = None
    if authenticated:
        stored_token = session["token"]
        public_data = {
            "access_key": stored_token["access_key"],
            "role": stored_token["role"],
            "status": stored_token.get("status"),
        }
    return web.json_response(
        {
            "authenticated": authenticated,
            "data": public_data,
            "session_id": session.identity,  # temporary wsproxy interop patch
        }
    )


async def login_handler(request: web.Request) -> web.Response:
    config = request.app["config"]
    session = await get_session(request)
    if session.get("authenticated", False):
        return web.HTTPBadRequest(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/generic-bad-request",
                    "title": "You have already logged in.",
                }
            ),
            content_type="application/problem+json",
        )
    request_headers = extra_config_headers.check(request.headers)
    secure_context = request_headers.get("X-BackendAI-Encoded", None)
    if not secure_context:
        # For non-encrypted requests, just read the body as-is.
        # Encrypted requests are handled by the `decrypt_payload` middleware.
        request["payload"] = await request.text()
    try:
        creds = json.loads(request["payload"])
    except json.JSONDecodeError as e:
        log.error("Login: JSON decoding error: {}", e)
        creds = {}
    if "username" not in creds or not creds["username"]:
        return web.HTTPBadRequest(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/invalid-api-params",
                    "title": "You must provide the username field.",
                }
            ),
            content_type="application/problem+json",
        )
    if "password" not in creds or not creds["password"]:
        return web.HTTPBadRequest(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/invalid-api-params",
                    "title": "You must provide the password field.",
                }
            ),
            content_type="application/problem+json",
        )
    result: MutableMapping[str, Any] = {
        "authenticated": False,
        "data": None,
    }

    async def _get_login_history():
        login_history = await request.app["redis"].get(
            f'login_history_{creds["username"]}',
        )
        if not login_history:
            login_history = {
                "last_login_attempt": 0,
                "login_fail_count": 0,
            }
        else:
            login_history = json.loads(login_history)
        if login_history["last_login_attempt"] < 0:
            login_history["last_login_attempt"] = 0
        if login_history["login_fail_count"] < 0:
            login_history["login_fail_count"] = 0
        return login_history

    async def _set_login_history(last_login_attempt, login_fail_count):
        """
        Set login history per email (not in browser session).
        """
        key = f'login_history_{creds["username"]}'
        value = json.dumps(
            {
                "last_login_attempt": last_login_attempt,
                "login_fail_count": login_fail_count,
            }
        )
        await request.app["redis"].set(key, value)

    # Block login if there are too many consecutive failed login attempts.
    BLOCK_TIME = config["session"]["login_block_time"]
    ALLOWED_FAIL_COUNT = config["session"]["login_allowed_fail_count"]
    login_time = time.time()
    login_history = await _get_login_history()
    last_login_attempt = login_history.get("last_login_attempt", 0)
    login_fail_count = login_history.get("login_fail_count", 0)
    if login_time - last_login_attempt > BLOCK_TIME:
        # If last attempt is far past, allow login again.
        login_fail_count = 0
    last_login_attempt = login_time
    if login_fail_count >= ALLOWED_FAIL_COUNT:
        log.info(
            "Too many consecutive login attempts for {}: {}",
            creds["username"],
            login_fail_count,
        )
        await _set_login_history(last_login_attempt, login_fail_count)
        return web.HTTPTooManyRequests(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/too-many-requests",
                    "title": "Too many failed login attempts",
                }
            ),
            content_type="application/problem+json",
        )

    try:
        anon_api_config = APIConfig(
            domain=config["api"]["domain"],
            endpoint=config["api"]["endpoint"][0],
            access_key="",
            secret_key="",  # anonymous session
            user_agent=user_agent,
            skip_sslcert_validation=not config["api"]["ssl_verify"],
        )
        assert anon_api_config.is_anonymous
        async with APISession(config=anon_api_config) as api_session:
            token = await api_session.User.authorize(creds["username"], creds["password"])
            stored_token = {
                "type": "keypair",
                "access_key": token.content["access_key"],
                "secret_key": token.content["secret_key"],
                "role": token.content["role"],
                "status": token.content.get("status"),
            }
            public_return = {
                "access_key": token.content["access_key"],
                "role": token.content["role"],
                "status": token.content.get("status"),
            }
            session["authenticated"] = True
            session["token"] = stored_token  # store full token
            result["authenticated"] = True
            result["data"] = public_return  # store public info from token
            login_fail_count = 0
            await _set_login_history(last_login_attempt, login_fail_count)
    except BackendClientError as e:
        # This is error, not failed login, so we should not update login history.
        return web.HTTPBadGateway(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/bad-gateway",
                    "title": "The proxy target server is inaccessible.",
                    "details": str(e),
                }
            ),
            content_type="application/problem+json",
        )
    except BackendAPIError as e:
        log.info("Authorization failed for {}: {}", creds["username"], e)
        result["authenticated"] = False
        result["data"] = {
            "type": e.data.get("type"),
            "title": e.data.get("title"),
            "details": e.data.get("msg"),
        }
        session["authenticated"] = False
        login_fail_count += 1
        await _set_login_history(last_login_attempt, login_fail_count)
    return web.json_response(result)


async def logout_handler(request: web.Request) -> web.Response:
    session = await get_session(request)
    session.invalidate()
    return web.Response(status=201)


async def webserver_healthcheck(_: web.Request) -> web.Response:
    result = {
        "version": __version__,
        "details": "Success",
    }
    return web.json_response(result)


async def token_login_handler(request: web.Request) -> web.Response:
    config = request.app["config"]

    # Check browser session exists.
    session = await get_session(request)
    if session.get("authenticated", False):
        return web.HTTPBadRequest(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/generic-bad-request",
                    "title": "You have already logged in.",
                }
            ),
            content_type="application/problem+json",
        )

    # Check if auth token is delivered through cookie.
    auth_token_name = config["api"]["auth_token_name"]
    auth_token = request.cookies.get(auth_token_name)
    if not auth_token:
        return web.HTTPBadRequest(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/invalid-api-params",
                    "title": "You must provide cookie-based authentication token",
                }
            ),
            content_type="application/problem+json",
        )

    # Login with the token.
    # We do not pose consecutive login failure for this handler since
    # user may frequently click edu-api launcher button.
    result: MutableMapping[str, Any] = {
        "authenticated": False,
        "data": None,
    }
    try:
        anon_api_config = APIConfig(
            domain=config["api"]["domain"],
            endpoint=config["api"]["endpoint"][0],
            access_key="",
            secret_key="",  # anonymous session
            user_agent=user_agent,
            skip_sslcert_validation=not config["api"]["ssl_verify"],
        )
        assert anon_api_config.is_anonymous
        async with APISession(config=anon_api_config) as api_session:
            # Send X-Forwarded-For header for token authentication with the client IP.
            client_ip = request.headers.get("X-Forwarded-For", request.remote)
            if client_ip:
                _headers = {"X-Forwarded-For": client_ip}
                api_session.aiohttp_session.headers.update(_headers)
            # Instead of email and password, cookie token will be used for auth.
            api_session.aiohttp_session.cookie_jar.update_cookies(request.cookies)
            token = await api_session.User.authorize("fake-email", "fake-pwd")
            stored_token = {
                "type": "keypair",
                "access_key": token.content["access_key"],
                "secret_key": token.content["secret_key"],
                "role": token.content["role"],
                "status": token.content.get("status"),
            }
            public_return = {
                "access_key": token.content["access_key"],
                "role": token.content["role"],
                "status": token.content.get("status"),
            }
            session["authenticated"] = True
            session["token"] = stored_token  # store full token
            result["authenticated"] = True
            result["data"] = public_return  # store public info from token
    except BackendClientError as e:
        return web.HTTPBadGateway(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/bad-gateway",
                    "title": "The proxy target server is inaccessible.",
                    "details": str(e),
                }
            ),
            content_type="application/problem+json",
        )
    except BackendAPIError as e:
        log.info("Authorization failed for token {}: {}", auth_token, e)
        result["authenticated"] = False
        result["data"] = {
            "type": e.data.get("type"),
            "title": e.data.get("title"),
            "details": e.data.get("msg"),
        }
        session["authenticated"] = False
    return web.json_response(result)


async def server_shutdown(app) -> None:
    pass


async def server_cleanup(app) -> None:
    await app["redis"].close()


@aiotools.server
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: Tuple[Any, ...],
) -> AsyncIterator[None]:
    config = args[0]
    app = web.Application(middlewares=[decrypt_payload])
    app["config"] = config
    j2env = jinja2.Environment(
        extensions=[
            "ai.backend.web.template.TOMLField",
            "ai.backend.web.template.TOMLStringListField",
        ],
        loader=jinja2.PackageLoader("ai.backend.web", "templates"),
    )
    j2env.filters["toml_scalar"] = toml_scalar
    app["j2env"] = j2env

    redis_url = yarl.URL("redis://host").with_host(config["session"]["redis"]["host"]).with_port(
        config["session"]["redis"]["port"]
    ).with_password(config["session"]["redis"]["password"]) / str(config["session"]["redis"]["db"])
    keepalive_options = {}
    if (_TCP_KEEPIDLE := getattr(socket, "TCP_KEEPIDLE", None)) is not None:
        keepalive_options[_TCP_KEEPIDLE] = 20
    if (_TCP_KEEPINTVL := getattr(socket, "TCP_KEEPINTVL", None)) is not None:
        keepalive_options[_TCP_KEEPINTVL] = 5
    if (_TCP_KEEPCNT := getattr(socket, "TCP_KEEPCNT", None)) is not None:
        keepalive_options[_TCP_KEEPCNT] = 3
    app["redis"] = await Redis.from_url(
        str(redis_url),
        socket_keepalive=True,
        socket_keepalive_options=keepalive_options,
    )

    if pidx == 0 and config["session"]["flush_on_startup"]:
        await app["redis"].flushdb()
        log.info("flushed session storage.")
    redis_storage = RedisStorage(
        app["redis"],
        max_age=config["session"]["max_age"],
    )

    setup_session(app, redis_storage)
    cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True, allow_methods="*", expose_headers="*", allow_headers="*"
        ),
    }
    cors = aiohttp_cors.setup(app, defaults=cors_options)

    anon_web_handler = partial(web_handler, is_anonymous=True)
    anon_web_plugin_handler = partial(web_plugin_handler, is_anonymous=True)

    app.router.add_route("HEAD", "/func/{path:folders/_/tus/upload/.*$}", anon_web_plugin_handler)
    app.router.add_route("PATCH", "/func/{path:folders/_/tus/upload/.*$}", anon_web_plugin_handler)
    app.router.add_route(
        "OPTIONS", "/func/{path:folders/_/tus/upload/.*$}", anon_web_plugin_handler
    )
    cors.add(app.router.add_route("POST", "/server/login", login_handler))
    cors.add(app.router.add_route("POST", "/server/token-login", token_login_handler))
    cors.add(app.router.add_route("POST", "/server/login-check", login_check_handler))
    cors.add(app.router.add_route("POST", "/server/logout", logout_handler))
    cors.add(app.router.add_route("GET", "/func/ping", webserver_healthcheck))
    cors.add(app.router.add_route("GET", "/func/{path:cloud/.*$}", anon_web_plugin_handler))
    cors.add(app.router.add_route("POST", "/func/{path:cloud/.*$}", anon_web_plugin_handler))
    cors.add(app.router.add_route("POST", "/func/{path:custom-auth/.*$}", anon_web_plugin_handler))
    cors.add(app.router.add_route("GET", "/func/{path:custom-auth/.*$}", anon_web_plugin_handler))
    cors.add(app.router.add_route("POST", "/func/{path:saml/.*$}", anon_web_plugin_handler))
    cors.add(app.router.add_route("POST", "/func/{path:auth/signup}", anon_web_plugin_handler))
    cors.add(app.router.add_route("POST", "/func/{path:auth/signout}", web_handler))
    cors.add(app.router.add_route("GET", "/func/{path:stream/kernel/_/events}", web_handler))
    cors.add(app.router.add_route("GET", "/func/{path:stream/session/[^/]+/apps$}", web_handler))
    cors.add(app.router.add_route("GET", "/func/{path:stream/.*$}", websocket_handler))
    cors.add(app.router.add_route("GET", "/func/", anon_web_handler))
    cors.add(app.router.add_route("HEAD", "/func/{path:.*$}", web_handler))
    cors.add(app.router.add_route("GET", "/func/{path:.*$}", web_handler))
    cors.add(app.router.add_route("PUT", "/func/{path:.*$}", web_handler))
    cors.add(app.router.add_route("POST", "/func/{path:.*$}", web_handler))
    cors.add(app.router.add_route("PATCH", "/func/{path:.*$}", web_handler))
    cors.add(app.router.add_route("DELETE", "/func/{path:.*$}", web_handler))
    cors.add(app.router.add_route("GET", "/pipeline/{path:.*$}", web_handler))
    cors.add(app.router.add_route("PUT", "/pipeline/{path:.*$}", web_handler))
    cors.add(app.router.add_route("POST", "/pipeline/{path:.*$}", web_handler))
    cors.add(app.router.add_route("PATCH", "/pipeline/{path:.*$}", web_handler))
    cors.add(app.router.add_route("DELETE", "/pipeline/{path:.*$}", web_handler))
    if config["service"]["mode"] == "webui":
        cors.add(app.router.add_route("GET", "/config.ini", config_ini_handler))
        cors.add(app.router.add_route("GET", "/config.toml", config_toml_handler))
        fallback_handler = console_handler
    elif config["service"]["mode"] == "static":
        fallback_handler = static_handler
    else:
        raise ValueError("Unrecognized service.mode", config["service"]["mode"])
    cors.add(app.router.add_route("GET", "/{path:.*$}", fallback_handler))

    app.on_shutdown.append(server_shutdown)
    app.on_cleanup.append(server_cleanup)

    async def on_prepare(request, response):
        # Remove "Server" header for a security reason.
        response.headers.popall("Server", None)

    app.on_response_prepare.append(on_prepare)

    ssl_ctx = None
    if config["service"]["ssl_enabled"]:
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(
            str(config["service"]["ssl_cert"]),
            str(config["service"]["ssl_privkey"]),
        )

    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(
        runner,
        str(config["service"]["ip"]),
        config["service"]["port"],
        backlog=1024,
        reuse_port=True,
        ssl_context=ssl_ctx,
    )
    await site.start()
    log.info("started.")

    try:
        yield
    finally:
        log.info("shutting down...")
        await runner.cleanup()


@click.command()
@click.option(
    "-f",
    "--config",
    "config_path",
    type=click.Path(exists=True),
    default="webserver.conf",
    help="The configuration file to use.",
)
@click.option("--debug", is_flag=True, default=False, help="Use more verbose logging.")
def main(config_path: str, debug: bool) -> None:
    raw_config = tomli.loads(Path(config_path).read_text(encoding="utf-8"))
    config = config_iv.check(raw_config)
    config["debug"] = debug
    if config["debug"]:
        debugFlag = "DEBUG"
    else:
        debugFlag = "INFO"
    setproctitle(f"backend.ai: webserver " f"{config['service']['ip']}:{config['service']['port']}")

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "colored": {
                    "()": "coloredlogs.ColoredFormatter",
                    "format": "%(asctime)s %(levelname)s %(name)s " "[%(process)d] %(message)s",
                    "field_styles": {
                        "levelname": {"color": 248, "bold": True},
                        "name": {"color": 246, "bold": False},
                        "process": {"color": "cyan"},
                        "asctime": {"color": 240},
                    },
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "DEBUG",
                    "formatter": "colored",
                    "stream": "ext://sys.stderr",
                },
                "null": {
                    "class": "logging.NullHandler",
                },
            },
            "loggers": {
                "": {
                    "handlers": ["console"],
                    "level": debugFlag,
                },
            },
        }
    )
    log.info("Backend.AI Web Server {0}", __version__)
    log.info("runtime: {0}", sys.prefix)
    log_config = logging.getLogger("ai.backend.web.config")
    log_config.debug("debug mode enabled.")
    if debug:
        print("== Web Server configuration ==")
        pprint(config)
    log.info("serving at {0}:{1}", config["service"]["ip"], config["service"]["port"])

    try:
        uvloop.install()
        aiotools.start_server(
            server_main,
            num_workers=min(4, os.cpu_count() or 1),
            args=(config,),
        )
    finally:
        log.info("terminated.")


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import base64
import json
import logging
import random
from typing import Optional, Tuple, Union, cast

import aiohttp
from aiohttp import web
from Crypto.Cipher import AES
from Crypto.Util.Padding import unpad

from ai.backend.client.exceptions import BackendAPIError, BackendClientError
from ai.backend.client.request import Request
from ai.backend.common.web.session import STORAGE_KEY, extra_config_headers, get_session

from .auth import get_anonymous_session, get_api_session
from .logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))

HTTP_HEADERS_TO_FORWARD = [
    "Accept-Language",
    "Authorization",
]


class WebSocketProxy:
    __slots__ = (
        "up_conn",
        "down_conn",
        "upstream_buffer",
        "upstream_buffer_task",
    )

    up_conn: aiohttp.ClientWebSocketResponse
    down_conn: web.WebSocketResponse
    upstream_buffer: asyncio.Queue[Tuple[Union[str, bytes], aiohttp.WSMsgType]]
    upstream_buffer_task: Optional[asyncio.Task]

    def __init__(
        self, up_conn: aiohttp.ClientWebSocketResponse, down_conn: web.WebSocketResponse
    ) -> None:
        self.up_conn = up_conn
        self.down_conn = down_conn
        self.upstream_buffer = asyncio.Queue()
        self.upstream_buffer_task = None

    async def proxy(self) -> None:
        asyncio.ensure_future(self.downstream())
        await self.upstream()

    async def upstream(self) -> None:
        try:
            async for msg in self.down_conn:
                if msg.type in (aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY):
                    await self.send(msg.data, msg.type)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    log.error(
                        "WebSocketProxy: connection closed with exception {}",
                        self.up_conn.exception(),
                    )
                    break
                elif msg.type == aiohttp.WSMsgType.CLOSE:
                    break
            # here, client gracefully disconnected
        except asyncio.CancelledError:
            # here, client forcibly disconnected
            pass
        finally:
            await self.close_downstream()

    async def downstream(self) -> None:
        try:
            self.upstream_buffer_task = asyncio.create_task(self.consume_upstream_buffer())
            async for msg in self.up_conn:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    await self.down_conn.send_str(msg.data)
                elif msg.type == aiohttp.WSMsgType.BINARY:
                    await self.down_conn.send_bytes(msg.data)
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    break
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    break
            # here, server gracefully disconnected
        except asyncio.CancelledError:
            pass
        except Exception as e:
            log.error("WebSocketProxy: unexpected error: {}", e)
        finally:
            await self.close_upstream()

    async def consume_upstream_buffer(self) -> None:
        try:
            while True:
                data, tp = await self.upstream_buffer.get()
                if not self.up_conn.closed:
                    if tp == aiohttp.WSMsgType.BINARY:
                        await self.up_conn.send_bytes(cast(bytes, data))
                    elif tp == aiohttp.WSMsgType.TEXT:
                        await self.up_conn.send_str(cast(str, data))
        except asyncio.CancelledError:
            pass

    async def send(self, msg: str, tp: aiohttp.WSMsgType) -> None:
        await self.upstream_buffer.put((msg, tp))

    async def close_downstream(self) -> None:
        if not self.down_conn.closed:
            await self.down_conn.close()

    async def close_upstream(self) -> None:
        if self.upstream_buffer_task is not None and not self.upstream_buffer_task.done():
            self.upstream_buffer_task.cancel()
            await self.upstream_buffer_task
        if not self.up_conn.closed:
            await self.up_conn.close()


@web.middleware
async def decrypt_payload(request: web.Request, handler) -> web.StreamResponse:
    request_headers = extra_config_headers.check(request.headers)
    secure_context = request_headers.get("X-BackendAI-Encoded", None)
    if secure_context:
        if not request.can_read_body:  # designated as encrypted but has an empty payload
            request["payload"] = ""
            return await handler(request)
        config = request.app["config"]
        scheme = config["service"]["force_endpoint_protocol"]
        if scheme is None:
            scheme = request.scheme
        api_endpoint = f"{scheme}://{request.host}"
        payload = await request.text()
        initial_vector, real_payload = payload.split(":")
        key = (base64.b64encode(api_endpoint.encode("ascii")).decode() + initial_vector * 2)[0:32]
        crypt = AES.new(
            bytes(key, encoding="utf8"), AES.MODE_CBC, bytes(initial_vector, encoding="utf8")
        )
        b64p = base64.b64decode(real_payload)
        request["payload"] = unpad(crypt.decrypt(bytes(b64p)), 16)
    else:
        # For all other requests without explicit encryption,
        # let the handler decide how to read the body.
        request["payload"] = ""
    return await handler(request)


async def web_handler(request, *, is_anonymous=False) -> web.StreamResponse:
    path = request.match_info.get("path", "")
    first_path = request.path.lstrip("/").partition("/")[
        0
    ]  # extract the first path  # extract the first path
    if is_anonymous:
        api_session = await asyncio.shield(get_anonymous_session(request))
    else:
        api_session = await asyncio.shield(get_api_session(request))
    if first_path == "pipeline":
        pipeline_endpoint = request.app["config"]["pipeline"]["endpoint"]
        api_session = await asyncio.shield(get_anonymous_session(request, pipeline_endpoint))
    try:
        async with api_session:
            # We perform request signing by ourselves using the HTTP session data,
            # but need to keep the client's version header so that
            # the final clients may perform its own API versioning support.
            request_headers = extra_config_headers.check(request.headers)
            request_api_version = request_headers.get("X-BackendAI-Version", None)
            secure_context = request_headers.get("X-BackendAI-Encoded", None)
            decrypted_payload_length = 0
            if secure_context:
                payload = request["payload"]
                decrypted_payload_length = len(payload)
            else:
                payload = request.content
            # Send X-Forwarded-For header for token authentication with the client IP.
            client_ip = request.headers.get("X-Forwarded-For")
            if not client_ip:
                client_ip = request.remote
            _headers = {"X-Forwarded-For": client_ip}
            api_session.aiohttp_session.headers.update(_headers)
            # Deliver cookie for token-based authentication.
            api_session.aiohttp_session.cookie_jar.update_cookies(request.cookies)
            # We treat all requests and responses as streaming universally
            # to be a transparent proxy.
            api_rqst = Request(
                request.method,
                path,
                payload,
                params=request.query,
                override_api_version=request_api_version,
            )
            if "Content-Type" in request.headers:
                api_rqst.content_type = request.content_type  # set for signing
                api_rqst.headers["Content-Type"] = request.headers[
                    "Content-Type"
                ]  # preserve raw value
            if "Content-Length" in request.headers and not secure_context:
                api_rqst.headers["Content-Length"] = request.headers["Content-Length"]
            if "Content-Length" in request.headers and secure_context:
                api_rqst.headers["Content-Length"] = str(decrypted_payload_length)
            for hdr in HTTP_HEADERS_TO_FORWARD:
                if request.headers.get(hdr) is not None:
                    api_rqst.headers[hdr] = request.headers[hdr]
            # Uploading request body happens at the entering of the block,
            # and downloading response body happens in the read loop inside.
            async with api_rqst.fetch() as up_resp:
                down_resp = web.StreamResponse()
                down_resp.set_status(up_resp.status, up_resp.reason)
                down_resp.headers.update(up_resp.headers)
                # We already have configured CORS handlers and the API server
                # also provides those headers.  Just let them as-is.
                await down_resp.prepare(request)
                while True:
                    chunk = await up_resp.read(8192)
                    if not chunk:
                        break
                    await down_resp.write(chunk)
                return down_resp
    except asyncio.CancelledError:
        raise
    except BackendAPIError as e:
        return web.Response(
            body=json.dumps(e.data),
            content_type="application/problem+json",
            status=e.status,
            reason=e.reason,
        )
    except BackendClientError:
        log.exception("web_handler: BackendClientError")
        return web.HTTPBadGateway(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/bad-gateway",
                    "title": "The proxy target server is inaccessible.",
                }
            ),
            content_type="application/problem+json",
        )
    except Exception:
        log.exception("web_handler: unexpected error")
        return web.HTTPInternalServerError(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/internal-server-error",
                    "title": "Something has gone wrong.",
                }
            ),
            content_type="application/problem+json",
        )
    finally:
        await api_session.close()


async def web_plugin_handler(request, *, is_anonymous=False) -> web.StreamResponse:
    """
    This handler is almost same to web_handler, but does not manipulate the
    content-type and content-length headers before sending up-requests.
    It also configures the domain in the json body for "auth/signup" requests.
    """
    path = request.match_info["path"]
    if is_anonymous:
        api_session = await asyncio.shield(get_anonymous_session(request))
    else:
        api_session = await asyncio.shield(get_api_session(request))
    try:
        async with api_session:
            content = request.content
            if path == "auth/signup":
                body = await request.json()
                body["domain"] = request.app["config"]["api"]["domain"]
                content = json.dumps(body).encode("utf8")
            request_api_version = request.headers.get("X-BackendAI-Version", None)
            # Send X-Forwarded-For header for token authentication with the client IP.
            client_ip = request.headers.get("X-Forwarded-For")
            if not client_ip:
                client_ip = request.remote
            _headers = {"X-Forwarded-For": client_ip}
            api_session.aiohttp_session.headers.update(_headers)
            # Deliver cookie for token-based authentication.
            api_session.aiohttp_session.cookie_jar.update_cookies(request.cookies)
            api_rqst = Request(
                request.method,
                path,
                content,
                params=request.query,
                content_type=request.content_type,
                override_api_version=request_api_version,
            )
            for hdr in HTTP_HEADERS_TO_FORWARD:
                if request.headers.get(hdr) is not None:
                    api_rqst.headers[hdr] = request.headers[hdr]
            async with api_rqst.fetch() as up_resp:
                down_resp = web.StreamResponse()
                down_resp.set_status(up_resp.status, up_resp.reason)
                down_resp.headers.update(up_resp.headers)
                # We already have configured CORS handlers and the API server
                # also provides those headers.  Just let them as-is.
                await down_resp.prepare(request)
                while True:
                    chunk = await up_resp.read(8192)
                    if not chunk:
                        break
                    await down_resp.write(chunk)
                return down_resp
    except asyncio.CancelledError:
        raise
    except BackendAPIError as e:
        return web.Response(
            body=json.dumps(e.data),
            content_type="application/problem+json",
            status=e.status,
            reason=e.reason,
        )
    except BackendClientError:
        log.exception("web_plugin_handler: BackendClientError")
        return web.HTTPBadGateway(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/bad-gateway",
                    "title": "The proxy target server is inaccessible.",
                }
            ),
            content_type="application/problem+json",
        )
    except Exception:
        log.exception("web_plugin_handler: unexpected error")
        return web.HTTPInternalServerError(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/internal-server-error",
                    "title": "Something has gone wrong.",
                }
            ),
            content_type="application/problem+json",
        )


async def websocket_handler(request, *, is_anonymous=False) -> web.StreamResponse:
    path = request.match_info["path"]
    session = await get_session(request)
    app = request.query.get("app")

    # Choose a specific Manager endpoint for persistent web app connection.
    api_endpoint = None
    should_save_session = False
    configured_endpoints = request.app["config"]["api"]["endpoint"]
    if session.get("api_endpoints", {}).get(app):
        stringified_endpoints = [str(e) for e in configured_endpoints]
        if session["api_endpoints"][app] in stringified_endpoints:
            api_endpoint = session["api_endpoints"][app]
    if api_endpoint is None:
        api_endpoint = random.choice(configured_endpoints)
        if "api_endpoints" not in session:
            session["api_endpoints"] = {}
        session["api_endpoints"][app] = str(api_endpoint)
        should_save_session = True

    if is_anonymous:
        api_session = await asyncio.shield(get_anonymous_session(request, api_endpoint))
    else:
        api_session = await asyncio.shield(get_api_session(request, api_endpoint))
    try:
        async with api_session:
            request_api_version = request.headers.get("X-BackendAI-Version", None)
            api_rqst = Request(
                request.method,
                path,
                request.content,
                params=request.query,
                content_type=request.content_type,
                override_api_version=request_api_version,
            )
            async with api_rqst.connect_websocket() as up_conn:
                down_conn = web.WebSocketResponse()
                await down_conn.prepare(request)
                web_socket_proxy = WebSocketProxy(up_conn.raw_websocket, down_conn)
                await web_socket_proxy.proxy()
                if should_save_session:
                    storage = request.get(STORAGE_KEY)
                    await storage.save_session(request, down_conn, session)
                return down_conn
    except asyncio.CancelledError:
        raise
    except BackendAPIError as e:
        return web.Response(
            body=json.dumps(e.data),
            content_type="application/problem+json",
            status=e.status,
            reason=e.reason,
        )
    except BackendClientError:
        log.exception("websocket_handler: BackendClientError")
        return web.HTTPBadGateway(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/bad-gateway",
                    "title": "The proxy target server is inaccessible.",
                }
            ),
            content_type="application/problem+json",
        )
    except Exception:
        log.exception("websocket_handler: unexpected error")
        return web.HTTPInternalServerError(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/internal-server-error",
                    "title": "Something has gone wrong.",
                }
            ),
            content_type="application/problem+json",
        )



import json
from typing import Any, List, Optional

from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.parser import Parser


class TOMLField(Extension):

    tags = {"toml_field"}

    def parse(self, parser: Parser) -> nodes.Node | List[nodes.Node]:
        tag_name = list(self.tags)[0]
        lineno = parser.stream.expect(f"name:{tag_name}").lineno
        field_name: nodes.Expr = parser.parse_expression()
        field_value: nodes.Expr = parser.parse_expression()
        return nodes.Output(
            [
                nodes.CondExpr(
                    # TOML does not have "null" syntax so let's comment out the field.
                    # We should skip "undefined" field also.
                    nodes.Or(
                        nodes.Test(field_value, "none", [], [], None, None),
                        nodes.Test(field_value, "undefined", [], [], None, None),
                    ),
                    nodes.Concat(
                        [
                            nodes.TemplateData("# "),
                            field_name,
                            nodes.TemplateData(" = "),
                        ],
                        lineno=lineno,
                    ),
                    nodes.Concat(
                        [
                            field_name,
                            nodes.TemplateData(" = "),
                            self._transform(field_value, lineno=lineno),
                        ],
                        lineno=lineno,
                    ),
                ),
            ],
            lineno=lineno,
        )

    def _transform(self, field_value: nodes.Expr, lineno: Optional[int] = None) -> nodes.Expr:
        field_value = nodes.Filter(field_value, "toml_scalar", [], [], None, None, lineno=lineno)
        return field_value


class TOMLStringListField(TOMLField):

    tags = {"toml_strlist_field"}

    def _transform(self, field_value: nodes.Expr, lineno: Optional[int] = None) -> nodes.Expr:
        field_value = nodes.Filter(
            field_value, "join", [nodes.Const(",")], [], None, None, lineno=lineno
        )
        field_value = nodes.Filter(field_value, "toml_scalar", [], [], None, None, lineno=lineno)
        return field_value


def toml_scalar(s: Any) -> str:
    """
    Encodes an arbitry Python object into a TOML-compatible string representation.
    The implementation borrows most of it from `json.dumps()`.

    Note that Jinja's default `tojson` filter uses its own HTML-safe unicode escaping for
    '<', '>', '&', "'" after applying `json.loads()` but we need to avoid doing so.

    ref) https://github.com/pallets/jinja/blob/7fb13bf/src/jinja2/utils.py#L657-L663
    """
    # If our custom tags are used, null values must be handled as commenting out the
    # entire field line and should not be passed to this filter.
    assert s is not None, "null is not allowed as a TOML scalar value"
    return json.dumps(s, ensure_ascii=False)



from pathlib import Path

import pkg_resources
import trafaret as t
import yarl

from ai.backend.common import validators as tx

default_static_path = Path(pkg_resources.resource_filename("ai.backend.web", "static")).resolve()

license_defs = {
    "edition": "Open Source",
    "valid_since": "",
    "valid_until": "",
}

config_iv = t.Dict(
    {
        t.Key("service"): t.Dict(
            {
                t.Key("ip", default="0.0.0.0"): tx.IPAddress,
                t.Key("port"): t.ToInt[1:65535],
                t.Key("wsproxy"): t.Dict(
                    {
                        t.Key("url", default=""): t.String(allow_blank=True),
                    }
                ),
                tx.AliasedKey(["ssl_enabled", "ssl-enabled"], default=False): t.ToBool,
                tx.AliasedKey(["ssl_cert", "ssl-cert"], default=None): t.Null
                | tx.Path(type="file"),
                tx.AliasedKey(["ssl_privkey", "ssl-privkey"], default=None): t.Null
                | tx.Path(type="file"),
                t.Key("static_path", default=default_static_path): tx.Path(type="dir"),
                tx.AliasedKey(
                    ["force_endpoint_protocol", "force-endpoint-protocol"],
                    default=None,
                ): t.Null
                | t.Enum("https", "http"),
                t.Key("mode"): t.Enum("webui", "static"),
                t.Key("enable_signup", default=False): t.ToBool,
                t.Key("allow_anonymous_change_password", default=False): t.ToBool,
                t.Key("allow_project_resource_monitor", default=False): t.ToBool,
                t.Key("allow_change_signin_mode", default=False): t.ToBool,
                t.Key("allow_manual_image_name_for_session", default=False): t.ToBool,
                t.Key("allow_signup_without_confirmation", default=False): t.ToBool,
                t.Key("always_enqueue_compute_session", default=False): t.ToBool,
                t.Key("webui_debug", default=False): t.ToBool,
                t.Key("mask_user_info", default=False): t.ToBool,
                t.Key("single_sign_on_vendors", default=None): t.Null
                | tx.StringList(empty_str_as_empty_list=True),
                t.Key("enable_container_commit", default=False): t.ToBool,
                t.Key("hide_agents", default=True): t.ToBool,
                t.Key("app_download_url", default=""): t.String(allow_blank=True),
            }
        ).allow_extra("*"),
        t.Key("resources"): t.Dict(
            {
                t.Key("open_port_to_public", default=False): t.ToBool,
                t.Key("max_cpu_cores_per_container", default=64): t.ToInt,
                t.Key("max_memory_per_container", default=64): t.ToInt,
                t.Key("max_cuda_devices_per_container", default=16): t.ToInt,
                t.Key("max_cuda_shares_per_container", default=16): t.ToInt,
                t.Key("max_shm_per_container", default=2): t.ToFloat,
                t.Key("max_file_upload_size", default=4294967296): t.ToInt,
            }
        ).allow_extra("*"),
        t.Key("environments"): t.Dict(
            {
                t.Key("allowlist", default=None): t.Null
                | tx.StringList(empty_str_as_empty_list=True),
            }
        ).allow_extra("*"),
        t.Key("plugin"): t.Dict(
            {
                t.Key("page", default=None): t.Null | tx.StringList(empty_str_as_empty_list=True),
            }
        ).allow_extra("*"),
        t.Key("pipeline"): t.Dict(
            {
                t.Key("endpoint", default=None): t.Null | tx.URL,
            }
        ).allow_extra("*"),
        t.Key("ui"): t.Dict(
            {
                t.Key("brand"): t.String,
                t.Key("default_environment", default=None): t.Null | t.String,
                t.Key("default_import_environment", default=None): t.Null | t.String,
                t.Key("menu_blocklist", default=None): t.Null
                | tx.StringList(empty_str_as_empty_list=True),
            }
        ).allow_extra("*"),
        t.Key("api"): t.Dict(
            {
                t.Key("domain"): t.String,
                t.Key("endpoint"): tx.DelimiterSeperatedList[yarl.URL](tx.URL, min_length=1),
                t.Key("text"): t.String,
                tx.AliasedKey(["ssl_verify", "ssl-verify"], default=True): t.ToBool,
                t.Key("auth_token_name", default="sToken"): t.String,
            }
        ).allow_extra("*"),
        t.Key("session"): t.Dict(
            {
                t.Key("redis"): t.Dict(
                    {
                        t.Key("host", default="localhost"): t.String,
                        t.Key("port", default=6379): t.ToInt[1:65535],
                        t.Key("db", default=0): t.ToInt,
                        t.Key("password", default=None): t.Null | t.String,
                    }
                ),
                t.Key("max_age", default=604800): t.ToInt,  # seconds (default: 1 week)
                t.Key("flush_on_startup", default=False): t.ToBool,
                t.Key("login_block_time", default=1200): t.ToInt,  # seconds (default: 20 min)
                t.Key("login_allowed_fail_count", default=10): t.ToInt,
                t.Key("auto_logout", default=False): t.ToBool,
            }
        ).allow_extra("*"),
        t.Key("license", default=license_defs): t.Dict(
            {
                t.Key("edition", default=license_defs["edition"]): t.String,
                t.Key("valid_since", default=license_defs["valid_since"]): t.String(
                    allow_blank=True
                ),
                t.Key("valid_until", default=license_defs["valid_until"]): t.String(
                    allow_blank=True
                ),
            }
        ).allow_extra("*"),
    }
).allow_extra("*")



import json
from typing import Optional

from aiohttp import web

from ai.backend.client.config import APIConfig
from ai.backend.client.session import AsyncSession as APISession
from ai.backend.common.web.session import get_session

from . import user_agent


async def get_api_session(
    request: web.Request,
    override_api_endpoint: Optional[str] = None,
) -> APISession:
    config = request.app["config"]
    api_endpoint = config["api"]["endpoint"][0]
    if override_api_endpoint is not None:
        api_endpoint = override_api_endpoint
    session = await get_session(request)
    if not session.get("authenticated", False):
        raise web.HTTPUnauthorized(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/auth-failed",
                    "title": "Unauthorized access",
                }
            ),
            content_type="application/problem+json",
        )
    if "token" not in session:
        raise web.HTTPUnauthorized(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/auth-failed",
                    "title": "Unauthorized access",
                }
            ),
            content_type="application/problem+json",
        )
    token = session["token"]
    if token["type"] != "keypair":
        raise web.HTTPBadRequest(
            text=json.dumps(
                {
                    "type": "https://api.backend.ai/probs/invalid-auth-params",
                    "title": "Incompatible auth token type.",
                }
            ),
            content_type="application/problem+json",
        )
    ak, sk = token["access_key"], token["secret_key"]
    api_config = APIConfig(
        domain=config["api"]["domain"],
        endpoint=api_endpoint,
        endpoint_type="api",
        access_key=ak,
        secret_key=sk,
        user_agent=user_agent,
        skip_sslcert_validation=not config["api"]["ssl_verify"],
    )
    return APISession(config=api_config, proxy_mode=True)


async def get_anonymous_session(
    request: web.Request,
    override_api_endpoint: Optional[str] = None,
) -> APISession:
    config = request.app["config"]
    api_endpoint = config["api"]["endpoint"][0]
    if override_api_endpoint is not None:
        api_endpoint = override_api_endpoint
    api_config = APIConfig(
        domain=config["api"]["domain"],
        endpoint=api_endpoint,
        endpoint_type="api",
        access_key="",
        secret_key="",
        user_agent=user_agent,
        skip_sslcert_validation=not config["api"]["ssl_verify"],
    )
    return APISession(config=api_config, proxy_mode=True)




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/web:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI WebUI Host',
    'install_requires': (
        'Jinja2~=3.1.2',
        'aiohttp_cors~=0.7',
        'aiohttp~=3.8.1',
        'aiotools~=1.5.9',
        """backend.ai-cli==22.09.7
""",
        """backend.ai-client==22.09.7
""",
        """backend.ai-common==22.09.7
""",
        """backend.ai-plugin==22.09.7
""",
        'click>=7.1.2',
        'coloredlogs~=15.0',
        'pycryptodome>=3.14.1',
        'redis[hiredis]~=4.3.4',
        'setproctitle~=1.2.2',
        'tomli~=2.0.1',
        'trafaret~=2.1',
        'types-click',
        'types-redis',
        'types-setuptools',
        'uvloop>=0.17',
        'yarl>=1.7',
    ),
    'license': 'LGPLv3',
    'long_description': """# Backend.AI Web Server

[![GitHub version](https://badge.fury.io/gh/lablup%2Fbackend.ai-webserver.svg)](https://badge.fury.io/gh/lablup%2Fbackend.ai-webserver) [![PyPI version](https://badge.fury.io/py/backend.ai-webserver.svg)](https://badge.fury.io/py/backend.ai-webserver)

A webapp hosting daemon which serves our `webui` as a SPA and proxies API requests


## Installation

Prepare a Python virtualenv (Python 3.9 or higher) and a Redis server (6.2 or higher).

```console
$ git clone https://github.com/lablup/backend.ai-webserver webserver
$ cd webserver
$ pip install -U -e .
$ cp webserver.sample.conf webserver.conf
```

## Mode

If `service.mode` is set "webui" (the default), the webserver handles
PWA-style fallbacks (e.g., serving `index.html` when there are no matching
files for the requested URL path).
The PWA must exclude `/server` and `/func` URL prefixes from its own routing
to work with the webserver's web sessions and the API proxy.

If it is set "static", the webserver serves the static files as-is,
without any fallbacks or hooking, while preserving the `/server` and `/func`
prefixed URLs and their functionalities.

If you want to serve web UI in webserver with "webui" mode, prepare static web UI source by choosing one of the followings.

### Option 1: Build web UI from source

Build **[backend.ai-webui](https://github.com/lablup/backend.ai-webui)** and copy all files under `build/bundle`
into the `src/ai/backend/web/static` directory.

### Option 2: Use pre-built web UI

To download and deploy web UI from pre-built source, do the following:

```console
git submodule init
git submodule update
cd src/ai/backend/web/static
git checkout main  # or target branch
git fetch
git pull
```
### Setup configuration for webserver

You don't have to write `config.toml` for the web UI as this webserver auto-generates it on-the-fly.

Edit `webserver.conf` to match with your environment.


## Usage

To execute web server, run command below. (for debugging, append a `--debug` flag)


```console
$ python -m ai.backend.web.server
```
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-webserver',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.web': (
            'VERSION',
            'py.typed',
            'static/.git',
            'static/README.md',
            'static/config.toml',
            'static/default.toml',
            'static/dist/components/backend-ai-agent-summary-view-8869df92.js',
            'static/dist/components/backend-ai-agent-view-1177b0d1.js',
            'static/dist/components/backend-ai-change-forgot-password-view-19427700.js',
            'static/dist/components/backend-ai-chart-e476f042.js',
            'static/dist/components/backend-ai-credential-view-5c0e3f61.js',
            'static/dist/components/backend-ai-data-view-3804d5fd.js',
            'static/dist/components/backend-ai-edu-applauncher-1cee54a6.js',
            'static/dist/components/backend-ai-email-verification-view-c4453e6b.js',
            'static/dist/components/backend-ai-environment-view-ef963a58.js',
            'static/dist/components/backend-ai-error-view-642433ea.js',
            'static/dist/components/backend-ai-import-view-a9b93cfb.js',
            'static/dist/components/backend-ai-information-view-c1a4d8a7.js',
            'static/dist/components/backend-ai-list-status-6c04082d.js',
            'static/dist/components/backend-ai-maintenance-view-0d8794e3.js',
            'static/dist/components/backend-ai-permission-denied-view-a1dfb9ac.js',
            'static/dist/components/backend-ai-session-launcher-9bc9dcd1.js',
            'static/dist/components/backend-ai-session-view-b89e0815.js',
            'static/dist/components/backend-ai-settings-view-1db5073c.js',
            'static/dist/components/backend-ai-statistics-view-c3695ddc.js',
            'static/dist/components/backend-ai-summary-view-4c4d83df.js',
            'static/dist/components/backend-ai-usersettings-view-90cfcc64.js',
            'static/dist/components/backend-ai-webui-765175f3.js',
            'static/dist/components/backend-ai-webui.js',
            'static/dist/components/chart-js-c7a002cd.js',
            'static/dist/components/dom-repeat-2fc0e1a5.js',
            'static/dist/components/expansion-e85b24e2.js',
            'static/dist/components/iconset-5c619972.js',
            'static/dist/components/input-behavior-9d643f20.js',
            'static/dist/components/json_to_csv-6fce0343.js',
            'static/dist/components/label-0b25f231.js',
            'static/dist/components/lablup-activity-panel-04bb9b1c.js',
            'static/dist/components/lablup-codemirror-36e88179.js',
            'static/dist/components/lablup-loading-spinner-553becce.js',
            'static/dist/components/lablup-progress-bar-5444537d.js',
            'static/dist/components/mwc-check-list-item-b05a927a.js',
            'static/dist/components/mwc-switch-185a66a2.js',
            'static/dist/components/mwc-tab-bar-693488b1.js',
            'static/dist/components/progress-spinner-52cba521.js',
            'static/dist/components/radio-behavior-9b90749e.js',
            'static/dist/components/select-35f6293f.js',
            'static/dist/components/slider-3c5ace6c.js',
            'static/dist/components/switch-e17613c9.js',
            'static/dist/components/tab-group-8f66a84b.js',
            'static/dist/components/textarea-5eaed463.js',
            'static/dist/components/textfield-3b465a9d.js',
            'static/dist/components/translate-unsafe-html-b8f4ae86.js',
            'static/dist/components/vaadin-grid-3d5eece2.js',
            'static/dist/components/vaadin-grid-filter-column-018e7d74.js',
            'static/dist/components/vaadin-grid-selection-column-10adf4bc.js',
            'static/dist/components/vaadin-grid-sort-column-a145f876.js',
            'static/dist/components/vaadin-item-2683cfb2.js',
            'static/dist/components/vaadin-item-styles-5b923f09.js',
            'static/index.html',
            'static/manifest.json',
            'static/manifest/app.js',
            'static/manifest/backend-ai.icns',
            'static/manifest/backend-ai.ico',
            'static/manifest/backend-ai.iconset/icon_128x128@1x.png',
            'static/manifest/backend-ai.iconset/icon_128x128@2x.png',
            'static/manifest/backend-ai.iconset/icon_16x16@1x.png',
            'static/manifest/backend-ai.iconset/icon_16x16@2x.png',
            'static/manifest/backend-ai.iconset/icon_256x256@1x.png',
            'static/manifest/backend-ai.iconset/icon_256x256@2x.png',
            'static/manifest/backend-ai.iconset/icon_32x32@1x.png',
            'static/manifest/backend-ai.iconset/icon_32x32@2x.png',
            'static/manifest/backend-ai.iconset/icon_512x512@1x.png',
            'static/manifest/backend-ai.iconset/icon_512x512@2x.png',
            'static/manifest/backend.ai-brand-simple.svg',
            'static/manifest/backend.ai-brand-white.svg',
            'static/manifest/backend.ai-brand.svg',
            'static/manifest/backend.ai-text.svg',
            'static/manifest/favicon.ico',
            'static/resources/app_template.json',
            'static/resources/badge-dark.svg',
            'static/resources/badge.svg',
            'static/resources/custom-kt.css',
            'static/resources/custom-original.css',
            'static/resources/custom.css',
            'static/resources/documents/privacy-policy.en.html',
            'static/resources/documents/privacy-policy.ko.html',
            'static/resources/documents/terms-of-service.en.html',
            'static/resources/documents/terms-of-service.ko.html',
            'static/resources/fashion-how-white.svg',
            'static/resources/fashion-how.svg',
            'static/resources/fonts/MaterialIcons-Regular.woff',
            'static/resources/fonts/MaterialIcons-Regular.woff2',
            'static/resources/fonts/MaterialIcons-Regular.woff2_',
            'static/resources/fonts/MaterialIcons-Regular.woff_',
            'static/resources/fonts/fa-brands-400.woff',
            'static/resources/fonts/fa-brands-400.woff2',
            'static/resources/fonts/fa-regular-400.woff',
            'static/resources/fonts/fa-regular-400.woff2',
            'static/resources/fonts/fa-solid-900.woff',
            'static/resources/fonts/fa-solid-900.woff2',
            'static/resources/fonts/font-awesome-all.min.css',
            'static/resources/fonts/font-awesome.css',
            'static/resources/fonts/fonts.css',
            'static/resources/fonts/montserrat-v13-latin-100.woff',
            'static/resources/fonts/montserrat-v13-latin-100.woff2',
            'static/resources/fonts/montserrat-v13-latin-100italic.woff',
            'static/resources/fonts/montserrat-v13-latin-100italic.woff2',
            'static/resources/fonts/montserrat-v13-latin-italic.woff',
            'static/resources/fonts/montserrat-v13-latin-italic.woff2',
            'static/resources/fonts/montserrat-v13-latin-regular.woff',
            'static/resources/fonts/montserrat-v13-latin-regular.woff2',
            'static/resources/fonts/quicksand-v13-latin-300.woff',
            'static/resources/fonts/quicksand-v13-latin-300.woff2',
            'static/resources/fonts/quicksand-v13-latin-500.woff',
            'static/resources/fonts/quicksand-v13-latin-500.woff2',
            'static/resources/fonts/quicksand-v13-latin-700.woff',
            'static/resources/fonts/quicksand-v13-latin-700.woff2',
            'static/resources/fonts/quicksand-v13-latin-regular.woff',
            'static/resources/fonts/quicksand-v13-latin-regular.woff2',
            'static/resources/fonts/roboto-v19-latin-100.woff',
            'static/resources/fonts/roboto-v19-latin-100.woff2',
            'static/resources/fonts/roboto-v19-latin-100italic.woff',
            'static/resources/fonts/roboto-v19-latin-100italic.woff2',
            'static/resources/fonts/roboto-v19-latin-300.woff',
            'static/resources/fonts/roboto-v19-latin-300.woff2',
            'static/resources/fonts/roboto-v19-latin-300italic.woff',
            'static/resources/fonts/roboto-v19-latin-300italic.woff2',
            'static/resources/fonts/roboto-v19-latin-900.woff',
            'static/resources/fonts/roboto-v19-latin-900.woff2',
            'static/resources/fonts/roboto-v19-latin-900italic.woff',
            'static/resources/fonts/roboto-v19-latin-900italic.woff2',
            'static/resources/fonts/roboto-v19-latin-italic.woff',
            'static/resources/fonts/roboto-v19-latin-italic.woff2',
            'static/resources/fonts/roboto-v19-latin-regular.woff',
            'static/resources/fonts/roboto-v19-latin-regular.woff2',
            'static/resources/fonts/ubuntu-v14-latin-300.woff',
            'static/resources/fonts/ubuntu-v14-latin-300.woff2',
            'static/resources/fonts/ubuntu-v14-latin-300italic.woff',
            'static/resources/fonts/ubuntu-v14-latin-300italic.woff2',
            'static/resources/fonts/ubuntu-v14-latin-500.woff',
            'static/resources/fonts/ubuntu-v14-latin-500.woff2',
            'static/resources/fonts/ubuntu-v14-latin-500italic.woff',
            'static/resources/fonts/ubuntu-v14-latin-500italic.woff2',
            'static/resources/fonts/ubuntu-v14-latin-700.woff',
            'static/resources/fonts/ubuntu-v14-latin-700.woff2',
            'static/resources/fonts/ubuntu-v14-latin-700italic.woff',
            'static/resources/fonts/ubuntu-v14-latin-700italic.woff2',
            'static/resources/fonts/ubuntu-v14-latin-italic.woff',
            'static/resources/fonts/ubuntu-v14-latin-italic.woff2',
            'static/resources/fonts/ubuntu-v14-latin-regular.woff',
            'static/resources/fonts/ubuntu-v14-latin-regular.woff2',
            'static/resources/i18n/de.json',
            'static/resources/i18n/el.json',
            'static/resources/i18n/en.json',
            'static/resources/i18n/es.json',
            'static/resources/i18n/fi.json',
            'static/resources/i18n/fr.json',
            'static/resources/i18n/id.json',
            'static/resources/i18n/it.json',
            'static/resources/i18n/ja.json',
            'static/resources/i18n/ko.json',
            'static/resources/i18n/mn.json',
            'static/resources/i18n/ms.json',
            'static/resources/i18n/pl.json',
            'static/resources/i18n/pt.json',
            'static/resources/i18n/pt_BR.json',
            'static/resources/i18n/ru.json',
            'static/resources/i18n/tr.json',
            'static/resources/i18n/vi.json',
            'static/resources/i18n/zh_CN.json',
            'static/resources/i18n/zh_TW.json',
            'static/resources/icons/LICENSE.md',
            'static/resources/icons/ROCm.png',
            'static/resources/icons/ROCm_black.png',
            'static/resources/icons/apache-spark.svg',
            'static/resources/icons/aws.png',
            'static/resources/icons/azure.png',
            'static/resources/icons/c.svg',
            'static/resources/icons/caffe.png',
            'static/resources/icons/caffe2.svg',
            'static/resources/icons/ceph.png',
            'static/resources/icons/check-list.svg',
            'static/resources/icons/cpp.svg',
            'static/resources/icons/default.png',
            'static/resources/icons/default_app.svg',
            'static/resources/icons/file_type_cuda.svg',
            'static/resources/icons/file_type_python.svg',
            'static/resources/icons/filebrowser.svg',
            'static/resources/icons/flux.png',
            'static/resources/icons/gcc.png',
            'static/resources/icons/gcp.png',
            'static/resources/icons/gdrive.png',
            'static/resources/icons/h2o.png',
            'static/resources/icons/haskell.png',
            'static/resources/icons/javascript.png',
            'static/resources/icons/julia.png',
            'static/resources/icons/jupyter.png',
            'static/resources/icons/jupyterlab.png',
            'static/resources/icons/lablup.png',
            'static/resources/icons/local.png',
            'static/resources/icons/lua.png',
            'static/resources/icons/matlab.png',
            'static/resources/icons/mlflow.png',
            'static/resources/icons/mxnet.svg',
            'static/resources/icons/nbp.png',
            'static/resources/icons/nni.png',
            'static/resources/icons/nni.svg',
            'static/resources/icons/nodejs.png',
            'static/resources/icons/novnc.svg',
            'static/resources/icons/nvidia.png',
            'static/resources/icons/octave.png',
            'static/resources/icons/openstack.png',
            'static/resources/icons/php.png',
            'static/resources/icons/purestorage.png',
            'static/resources/icons/python.png',
            'static/resources/icons/pytorch.svg',
            'static/resources/icons/r-lang.svg',
            'static/resources/icons/rust.svg',
            'static/resources/icons/sftp.png',
            'static/resources/icons/sftp2.png',
            'static/resources/icons/swift-tensorflow.png',
            'static/resources/icons/swift.svg',
            'static/resources/icons/tensorflow.png',
            'static/resources/icons/terminal.svg',
            'static/resources/icons/texlive.png',
            'static/resources/icons/tpu.svg',
            'static/resources/icons/ubuntu.svg',
            'static/resources/icons/vscode.svg',
            'static/resources/icons/xfce.png',
            'static/resources/icons/xrdp.svg',
            'static/resources/image_metadata.json',
            'static/resources/images/401_unauthorized_access.svg',
            'static/resources/images/404_not_found.svg',
            'static/resources/images/launcher-background.png',
            'static/resources/images/loading-background-large.jpg',
            'static/resources/images/loading-background.jpg',
            'static/resources/images/web-terminal-guide-1.png',
            'static/resources/images/web-terminal-guide-2.png',
            'static/resources/images/web-terminal-guide-3.png',
            'static/resources/images/web-terminal-guide-4.png',
            'static/resources/logo.png',
            'static/resources/menu_icons/icon_create_a_key_pair.svg',
            'static/resources/menu_icons/icon_keypair_management.svg',
            'static/resources/menu_icons/icon_upload_files.svg',
            'static/resources/storage_metadata.json',
            'static/resources/templates/under_construction.html',
            'static/src/lib/bundles/webcomponents-ce.js',
            'static/src/lib/bundles/webcomponents-ce.js.map',
            'static/src/lib/bundles/webcomponents-pf_dom.js',
            'static/src/lib/bundles/webcomponents-pf_dom.js.map',
            'static/src/lib/bundles/webcomponents-pf_js.js',
            'static/src/lib/bundles/webcomponents-pf_js.js.map',
            'static/src/lib/bundles/webcomponents-sd-ce-pf.js',
            'static/src/lib/bundles/webcomponents-sd-ce-pf.js.map',
            'static/src/lib/bundles/webcomponents-sd-ce.js',
            'static/src/lib/bundles/webcomponents-sd-ce.js.map',
            'static/src/lib/bundles/webcomponents-sd.js',
            'static/src/lib/bundles/webcomponents-sd.js.map',
            'static/src/lib/web-animations-js/web-animations-next-lite.min.js',
            'static/src/lib/webcomponents-loader.js',
            'static/sw.js',
            'static/sw.js.map',
            'static/version.json',
            'static/workbox-6da860f9.js',
            'static/workbox-6da860f9.js.map',
            'templates/config.toml.j2',
            'templates/config_ini.toml.j2',
        ),
    },
    'packages': (
        'ai.backend.web',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.10,<3.11',
    'url': 'https://github.com/lablup/backend.ai',
    'version': """22.09.7
""",
    'zip_safe': False,
})
